/*
 *  osconf.c - ACPI OS-dependent functions for Kernel Boot/Configuration time
 *
 *  Copyright (C) 2000 Intel Corp.
 *  Copyright (C) 2000 J.I. Lee <Jung-Ik.Lee@intel.com>
 *
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/bootmem.h>
#include <linux/pci.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/sal.h>

#include "acpi.h"
#include "osconf.h"


static void * __init acpi_os_allocate_bt(u32 size);
static void * __init acpi_os_callocate_bt(u32 size);
static void __init acpi_os_free_bt(void *ptr);

static ACPI_STATUS __init
acpi_os_queue_for_execution_bt(
	u32			priority,
	OSD_EXECUTION_CALLBACK	callback,
	void			*context
	);

static ACPI_STATUS __init
acpi_os_read_pci_cfg_byte_bt( u32 segbus, u32 func, u32 addr, u8 * val);

static ACPI_STATUS __init
acpi_os_read_pci_cfg_word_bt( u32 segbus, u32 func, u32 addr, u16 * val);

static ACPI_STATUS __init
acpi_os_read_pci_cfg_dword_bt( u32 segbus, u32 func, u32 addr, u32 * val);

static ACPI_STATUS __init
acpi_os_write_pci_cfg_byte_bt( u32 segbus, u32 func, u32 addr, u8 val);

static ACPI_STATUS __init
acpi_os_write_pci_cfg_word_bt( u32 segbus, u32 func, u32 addr, u16 val);

static ACPI_STATUS __init
acpi_os_write_pci_cfg_dword_bt( u32 segbus, u32 func, u32 addr, u32 val);


static struct acpi_osd *acpi_osd;
extern struct acpi_osd acpi_osd_rt;
static struct acpi_osd acpi_osd_bt __initdata = {
	/* these are boottime osd entries that differ from runtime entries */
	acpi_os_allocate_bt,
	acpi_os_callocate_bt,
	acpi_os_free_bt,
	acpi_os_queue_for_execution_bt,
	acpi_os_read_pci_cfg_byte_bt,
	acpi_os_read_pci_cfg_word_bt,
	acpi_os_read_pci_cfg_dword_bt,
	acpi_os_write_pci_cfg_byte_bt,
	acpi_os_write_pci_cfg_word_bt,
	acpi_os_write_pci_cfg_dword_bt
};

#ifdef	CONFIG_ACPI_KERNEL_CONFIG_BM_PROFILE
static void	__init
acpi_cf_bm_statistics( void );
#endif

void __init
acpi_os_bind_osd(int	acpi_phase)
{
	switch (acpi_phase) {
	case	ACPI_CF_PHASE_BOOTTIME:
		acpi_osd = &acpi_osd_bt;
		printk("Acpi cfg:bind to Boot time Acpi OSD\n");
		break;
	case	ACPI_CF_PHASE_RUNTIME:
	default:
		acpi_osd = &acpi_osd_rt;
		printk("Acpi cfg:bind to Run time Acpi OSD\n");
#ifdef	CONFIG_ACPI_KERNEL_CONFIG_BM_PROFILE
		acpi_cf_bm_statistics();
#endif
		break;
	}
}

void *
acpi_os_allocate(u32 size)
{
	return acpi_osd->allocate(size);
}

void *
acpi_os_callocate(u32 size)
{
	return acpi_osd->callocate(size);
}

void
acpi_os_free(void *ptr)
{
	acpi_osd->free(ptr);
	return;
}


ACPI_STATUS
acpi_os_read_pci_cfg_byte( u32 segbus, u32 func, u32 addr, u8 * val)
{
	return acpi_osd->read_pci_cfg_byte(segbus, func, addr, val);
}


ACPI_STATUS
acpi_os_read_pci_cfg_word( u32 segbus, u32 func, u32 addr, u16 * val)
{
	return acpi_osd->read_pci_cfg_word(segbus, func, addr, val);
}


ACPI_STATUS
acpi_os_read_pci_cfg_dword( u32 segbus, u32 func, u32 addr, u32 * val)
{
	return acpi_osd->read_pci_cfg_dword(segbus, func, addr, val);
}


ACPI_STATUS
acpi_os_write_pci_cfg_byte( u32 segbus, u32 func, u32 addr, u8 val)
{
	return acpi_osd->write_pci_cfg_byte(segbus, func, addr, val);
}


ACPI_STATUS
acpi_os_write_pci_cfg_word( u32 segbus, u32 func, u32 addr, u16 val)
{
	return acpi_osd->write_pci_cfg_word(segbus, func, addr, val);
}


ACPI_STATUS
acpi_os_write_pci_cfg_dword( u32 segbus, u32 func, u32 addr, u32 val)
{
	return acpi_osd->write_pci_cfg_dword(segbus, func, addr, val);
}

#ifdef	CONFIG_ACPI_KERNEL_CONFIG_BM_PROFILE
/*
 * Let's profile bootmem usage to see how much we consume. J.I.
 */
static	unsigned long	bm_alloc_size __initdata = 0;
static	unsigned long	bm_alloc_size_max __initdata = 0;
static	unsigned long	bm_alloc_count_max __initdata = 0;
static	unsigned long	bm_free_count_max __initdata = 0;

static void	__init
acpi_cf_bm_checkin(void *ptr, u32 size)
{
	bm_alloc_count_max++;
	bm_alloc_size += size;
	if (bm_alloc_size > bm_alloc_size_max)
		bm_alloc_size_max = bm_alloc_size;
};

static void	__init
acpi_cf_bm_checkout(void *ptr, u32 size)
{
	bm_free_count_max++;
	bm_alloc_size -= size;
};

static void	__init
acpi_cf_bm_statistics( void )
{
	printk("Acpi cfg:bm_alloc_size_max =%ld bytes\n", bm_alloc_size_max);
	printk("Acpi cfg:bm_alloc_count_max=%ld\n", bm_alloc_count_max);
	printk("Acpi cfg:bm_free_count_max =%ld\n", bm_free_count_max);
}
#endif


static void * __init
acpi_os_allocate_bt(u32 size)
{
	void *ptr;

	size += sizeof(unsigned long);
	ptr = alloc_bootmem(size);

	if (ptr) {
#ifdef	CONFIG_ACPI_KERNEL_CONFIG_BM_PROFILE
		acpi_cf_bm_checkin(ptr, size);
#endif
		*((unsigned long *)ptr) = (unsigned long)size;
		ptr += sizeof(unsigned long);
	}

	return ptr;
}

static void * __init
acpi_os_callocate_bt(u32 size)
{
	void *ptr = acpi_os_allocate_bt(size);

	return ptr;
}

static void __init
acpi_os_free_bt(void *ptr)
{
	unsigned long	size;

	ptr -= sizeof(size);
	size = *((unsigned long *)ptr);

#ifdef	CONFIG_ACPI_KERNEL_CONFIG_BM_PROFILE
	acpi_cf_bm_checkout(ptr, (unsigned long)size);
#endif
	//if (size)
		free_bootmem (__pa((unsigned long)ptr), (u32)size);
}


static ACPI_STATUS __init
acpi_os_queue_for_execution_bt(
	u32			priority,
	OSD_EXECUTION_CALLBACK	callback,
	void			*context)
{
	/*
	 * run callback immediately
	 */
	(*callback)(context);
	return AE_OK;
}


static ACPI_STATUS __init
acpi_os_read_pci_cfg_byte_bt( u32 segbus, u32 func, u32 addr, u8 * val)
{
	unsigned int	devfn;
	s64		status;
	u64		lval;

	devfn = PCI_DEVFN((func >> 16) & 0xffff, func & 0xffff);
	status = ia64_sal_pci_config_read(PCI_CONFIG_ADDRESS((segbus & 0xffff), devfn, addr), 1, &lval);
	*val = lval;

	return status;
}


static ACPI_STATUS __init
acpi_os_read_pci_cfg_word_bt( u32 segbus, u32 func, u32 addr, u16 * val)
{
	unsigned int	devfn;
	s64		status;
	u64		lval;

	devfn = PCI_DEVFN((func >> 16) & 0xffff, func & 0xffff);
	status = ia64_sal_pci_config_read(PCI_CONFIG_ADDRESS((segbus & 0xffff), devfn, addr), 2, &lval);
	*val = lval;

	return status;
}


static ACPI_STATUS __init
acpi_os_read_pci_cfg_dword_bt( u32 segbus, u32 func, u32 addr, u32 * val)
{
	unsigned int	devfn;
	s64		status;
	u64		lval;

	devfn = PCI_DEVFN((func >> 16) & 0xffff, func & 0xffff);
	status = ia64_sal_pci_config_read(PCI_CONFIG_ADDRESS((segbus & 0xffff), devfn, addr), 4, &lval);
	*val = lval;

	return status;
}


static ACPI_STATUS __init
acpi_os_write_pci_cfg_byte_bt( u32 segbus, u32 func, u32 addr, u8 val)
{
	unsigned int	devfn;

	devfn = PCI_DEVFN((func >> 16) & 0xffff, func & 0xffff);
	return ia64_sal_pci_config_write(PCI_CONFIG_ADDRESS((segbus & 0xffff), devfn, addr), 1, val);
}


static ACPI_STATUS __init
acpi_os_write_pci_cfg_word_bt( u32 segbus, u32 func, u32 addr, u16 val)
{
	unsigned int	devfn;

	devfn = PCI_DEVFN((func >> 16) & 0xffff, func & 0xffff);
	return ia64_sal_pci_config_write(PCI_CONFIG_ADDRESS((segbus & 0xffff), devfn, addr), 2, val);
}


static ACPI_STATUS __init
acpi_os_write_pci_cfg_dword_bt( u32 segbus, u32 func, u32 addr, u32 val)
{
	unsigned int	devfn;

	devfn = PCI_DEVFN((func >> 16) & 0xffff, func & 0xffff);
	return ia64_sal_pci_config_write(PCI_CONFIG_ADDRESS((segbus & 0xffff), devfn, addr), 4, val);
}


