/************************************************************
 * EFI GUID Partition Table handling
 * Per Intel EFI Specification v0.99
 * http://developer.intel.com/technology/efi/efi.htm
 * efi.[ch] by Matt Domsch <Matt_Domsch@dell.com>
 *   Copyright 2000 Dell Computer Corporation
 *
 * Note, the EFI Specification, v0.99, has a reference to
 * Dr. Dobbs Journal, May 1994 (actually it's in May 1992)
 * but that isn't the CRC function being used by EFI.  Intel's
 * EFI Sample Implementation shows that they use the same function
 * as was COPYRIGHT (C) 1986 Gary S. Brown.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 * TODO:
 *
 * Changelog:
 * Tue Dec  5 2000 Matt Domsch <Matt_Domsch@dell.com>
 * - Moved crc32() to linux/lib, added efi_crc32().
 *
 * Thu Nov 30 2000 Matt Domsch <Matt_Domsch@dell.com>
 * - Replaced Intel's CRC32 function with an equivalent
 *   non-license-restricted version.
 *
 * Wed Oct 25 2000 Matt Domsch <Matt_Domsch@dell.com>
 * - Fixed the LastLBA() call to return the proper last block
 *
 * Thu Oct 12 2000 Matt Domsch <Matt_Domsch@dell.com>
 * - Thanks to Andries Brouwer for his debugging assistance.
 * - Code works, detects all the partitions.
 *
 ************************************************************/
#include <linux/config.h>
#include <linux/fs.h>
#include <linux/genhd.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/blk.h>
#include <linux/blkpg.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <asm/system.h>
#include <asm/efi.h>
#include <linux/crc32.h>

#include "check.h"
#include "efi.h"

#if CONFIG_BLK_DEV_MD
extern void md_autodetect_dev(kdev_t dev);
#endif


#undef EFI_DEBUG
#ifdef EFI_DEBUG
static char *efi_printk_level = KERN_DEBUG;
#define debug_printk printk
#else
#define debug_printk(...)
#endif

/************************************************************
 * efi_crc32()
 * Requires:
 *  - a buffer of length len
 * Modifies: nothing
 * Returns:
 *  EFI-style CRC32 value for buf
 *
 * This function uses the crc32 function by Gary S. Brown,
 * but seeds the function with ~0, and xor's with ~0 at the end.
 ************************************************************/

u32 static inline
efi_crc32(const void *buf, unsigned long len)
{
  return (crc32(buf, len, ~0L) ^ ~0L);
}




/************************************************************
 * IsLegacyMBRValid()
 * Requires:
 *  - mbr is a pointer to a legacy mbr structure
 * Modifies: nothing
 * Returns:
 *  1 on true
 *  0 on false
 ************************************************************/
static inline int
IsLegacyMBRValid(LegacyMBR_t *mbr)
{
	return (mbr ? (mbr->Signature == MSDOS_MBR_SIGNATURE) : 0);
}



/************************************************************
 * LastLBA()
 * Requires:
 *  - struct gendisk hd
 *  - kdev_t dev
 * Modifies: nothing
 * Returns:
 *  Last LBA value on success.  This is stored (by sd and
 *  ide-geometry) in
 *  the part[0] entry for this disk, and is the number of
 *  physical sectors available on the disk.
 *  0 on error
 ************************************************************/
u64
LastLBA(struct gendisk *hd, kdev_t dev)
{
  if (!hd || !hd->part) return 0;
  return hd->part[MINOR(dev)].nr_sects - 1;
}


/************************************************************
 * ReadLBA()
 * Requires:
 *  - hd is our disk device.
 *  - dev is our device major number
 *  - lba is the logical block address desired (disk hardsector number)
 *  - buffer is a buffer of size size into which data copied
 *  - size_t count is size of the read (in bytes)
 * Modifies:
 *  - buffer
 * Returns:
 *  - count of bytes read
 *  - 0 on error
 * Bugs:
 *  - bread() takes second argument as a signed int, not a u64.
 *    This is because getblk() takes the block number as a signed int.
 *    This overflow is known on l-k.   We overflow at about 1TB.
 *
 ************************************************************/

static size_t
ReadLBA(struct gendisk *hd, kdev_t dev, u64 _lba, u8 *buffer, size_t count)
{
	struct buffer_head *bh;
	size_t totalreadcount = 0, bytesread;
	int lba = (_lba & 0x7FFFFFFF), i, blockstoread, blocksize;
	debug_printk(efi_printk_level "ReadLBA(%p,%s,%x,%p,%x)\n",
		     hd, kdevname(dev), lba, buffer, count);

	if (!hd || !buffer || !count) return 0;


	blocksize = get_hardsect_size(dev);
	blockstoread = count / blocksize;
	if (count % blocksize) blockstoread += 1;
	debug_printk(efi_printk_level "about to read %d blocks\n",
		     blockstoread);


	for (i=0; i<blockstoread; i++) {
		bh = bread(dev, lba+i, blocksize);
		if (!bh) {
			/* We hit the end of the disk */
			debug_printk(efi_printk_level
				     "bread returned NULL.\n");
			return totalreadcount;
		}

		bytesread = (count > bh->b_size ? bh->b_size : count);
		memcpy(buffer, bh->b_data, bytesread);

		buffer += bytesread;         /* Advance the buffer pointer */
		totalreadcount += bytesread; /* Advance the total read count */
		count -= bytesread;         /* Subtract bytesread from count */

		brelse(bh);
	}

	return totalreadcount;
}

void
PrintGuidPartitionTableHeader(GuidPartitionTableHeader_t *gpt)
{
	debug_printk(efi_printk_level "GUID Partition Table Header\n");
	if (!gpt) return;
	debug_printk(efi_printk_level "Signature      : %lx\n",
		     gpt->Signature);
	debug_printk(efi_printk_level "Revision       : %x\n",
		     gpt->Revision);
	debug_printk(efi_printk_level "HeaderSize     : %x\n",
		     gpt->HeaderSize);
	debug_printk(efi_printk_level "HeaderCRC32    : %x\n",
		     gpt->HeaderCRC32);
	debug_printk(efi_printk_level "MyLBA          : %lx\n",
		     gpt->MyLBA);
	debug_printk(efi_printk_level "AlternateLBA   : %lx\n",
		     gpt->AlternateLBA);
	debug_printk(efi_printk_level "FirstUsableLBA : %lx\n",
		     gpt->FirstUsableLBA);
	debug_printk(efi_printk_level "LastUsableLBA  : %lx\n",
		     gpt->LastUsableLBA);

	debug_printk(efi_printk_level "PartitionEntryLBA : %lx\n",
		     gpt->PartitionEntryLBA);
	debug_printk(efi_printk_level "NumberOfPartitionEntries : %x\n",
		     gpt->NumberOfPartitionEntries);
	debug_printk(efi_printk_level "SizeOfPartitionEntry : %x\n",
		     gpt->SizeOfPartitionEntry);
	debug_printk(efi_printk_level "PartitionEntryArrayCRC32 : %x\n",
		     gpt->PartitionEntryArrayCRC32);

	return;
}



/************************************************************
 * ReadGuidPartitionEntries()
 * Requires:
 *  - filedes is an open file descriptor, suitable for reading
 *  - lba is the Logical Block Address of the partition table
 *  - gpt is a buffer into which the GPT will be put
 * Modifies:
 *  - filedes file and pointer
 *  - gpt
 * Returns:
 *   pte on success
 *   NULL on error
 * Notes: remember to free pte when you're done!
 ************************************************************/
GuidPartitionEntry_t *
ReadGuidPartitionEntries(struct gendisk *hd, kdev_t dev,
			 GuidPartitionTableHeader_t *gpt)
{
	size_t count;
	GuidPartitionEntry_t *pte;
	if (!hd || !gpt) return NULL;

	count = gpt->NumberOfPartitionEntries * gpt->SizeOfPartitionEntry;
	debug_printk(efi_printk_level "ReadGPTEs() kmallocing %x bytes\n",
		     count);
	if (!count) return NULL;
	pte = kmalloc(count, GFP_KERNEL);
	if (!pte)  return NULL;
	memset(pte, 0, count);

	if (ReadLBA(hd, dev, gpt->PartitionEntryLBA, (u8 *)pte,
		    count) < count) {
		kfree(pte);
		return NULL;
	}
	return pte;
}



/************************************************************
 * ReadGuidPartitionTableHeader()
 * Requires:
 *  - hd is our struct gendisk
 *  - dev is our device major number
 *  - lba is the Logical Block Address of the partition table
 *  - gpt is a buffer into which the GPT will be put
 *  - pte is a buffer into which the PTEs will be put
 * Modifies:
 *  - gpt and pte
 * Returns:
 *   1 on success
 *   0 on error
 ************************************************************/

GuidPartitionTableHeader_t *
ReadGuidPartitionTableHeader(struct gendisk *hd, kdev_t dev, u64 lba)

{
	GuidPartitionTableHeader_t *gpt;
	if (!hd) return NULL;

	gpt = kmalloc(sizeof(GuidPartitionTableHeader_t), GFP_KERNEL);
	if (!gpt) return NULL;
	memset(gpt, 0, sizeof(GuidPartitionTableHeader_t));

	debug_printk(efi_printk_level "GPTH() calling ReadLBA().\n");
	if (ReadLBA(hd, dev, lba, (u8 *)gpt,
		    sizeof(GuidPartitionTableHeader_t)) <
	    sizeof(GuidPartitionTableHeader_t)) {
		debug_printk(efi_printk_level "ReadGPTH(%lx) read failed.\n",
			     lba);
		kfree(gpt);
		return NULL;
	}
	PrintGuidPartitionTableHeader(gpt);

	return gpt;
}



/************************************************************
 * IsGuidPartitionTableValid()
 * Requires:
 *  - gd points to our struct gendisk
 *  - dev is our device major number
 *  - lba is the logical block address of the GPTH to test
 *  - gpt is a GPTH if it's valid
 *  - ptes is a PTEs if it's valid
 * Modifies:
 *  - gpt and ptes
 * Returns:
 *   1 if valid
 *   0 on error
 ************************************************************/
static int
IsGuidPartitionTableValid(struct gendisk *hd, kdev_t dev, u64 lba,
			  GuidPartitionTableHeader_t **gpt,
			  GuidPartitionEntry_t **ptes)
{
	u32 crc, origcrc;

	if (!hd || !gpt || !ptes) return 0;
	if (!(*gpt = ReadGuidPartitionTableHeader(hd, dev, lba))) return 0;

	/* Check the GUID Partition Table Signature */
	if ((*gpt)->Signature != GUID_PT_HEADER_SIGNATURE) {
		debug_printk(efi_printk_level "GUID Partition Table Header Signature is wrong: %x != %x\n", (*gpt)->Signature, GUID_PT_HEADER_SIGNATURE);
		kfree(*gpt);
		*gpt = NULL;
		return 0;
	}

	/* Check the GUID Partition Table CRC */
	origcrc = (*gpt)->HeaderCRC32;
	(*gpt)->HeaderCRC32 = 0;
	crc = efi_crc32((const unsigned char *)(*gpt), (*gpt)->HeaderSize);


	if (crc != origcrc) {
		debug_printk(efi_printk_level "GUID Partition Table Header CRC is wrong: %x != %x\n", (*gpt)->HeaderCRC32, origcrc);
		kfree(*gpt);
		*gpt = NULL;
		return 0;
	}
	(*gpt)->HeaderCRC32 = origcrc;

	/* Check that the MyLBA entry points to the LBA that contains
	 * the GUID Partition Table */
	if ((*gpt)->MyLBA != lba) {
		debug_printk(efi_printk_level "GPT MyLBA incorrect: %lx != %lx\n", (*gpt)->MyLBA, lba);
		kfree(*gpt);
		*gpt = NULL;
		return 0;
	}

	if (!(*ptes = ReadGuidPartitionEntries(hd, dev, *gpt))) {
		debug_printk(efi_printk_level "read PTEs failed.\n");
		kfree(*gpt);
		*gpt = NULL;
		return 0;
	}

	/* Check the GUID Partition Entry Array CRC */
	crc = efi_crc32((const unsigned char *)(*ptes),
		    (*gpt)->NumberOfPartitionEntries *
			   (*gpt)->SizeOfPartitionEntry);

	if (crc != (*gpt)->PartitionEntryArrayCRC32)  {
		debug_printk(efi_printk_level "GUID Partitition Entry Array CRC check failed.\n");
		kfree(*gpt);
		*gpt = NULL;
		kfree(*ptes);
		*ptes = NULL;
		return 0;
	}


	/* We're done, all's well */
	return 1;
}



/************************************************************
 * FindValidGPT()
 * Requires:
 *  - gd points to our struct gendisk
 *  - dev is our device major number
 *  - gpt is a GPTH if it's valid
 *  - ptes is a PTE
 * Modifies:
 *  - gpt & ptes
 * Returns:
 *   1 if valid
 *   0 on error
 ************************************************************/
static int
FindValidGPT(struct gendisk *hd, kdev_t dev,
	     GuidPartitionTableHeader_t **gpt,
	     GuidPartitionEntry_t **ptes)
{
	int rc = 0;
	GuidPartitionTableHeader_t *pgpt = NULL, *agpt = NULL;
	GuidPartitionEntry_t *pptes = NULL, *aptes = NULL;
	u64 lastlba;
	if (!hd || !gpt || !ptes) return 0;

	lastlba = LastLBA(hd, dev);
	/* Check the Primary GPT */
	rc = IsGuidPartitionTableValid(hd, dev, 1, &pgpt, &pptes);
	if (rc) {
		/* Primary GPT is OK, check the alternate and warn if bad */
		rc = IsGuidPartitionTableValid(hd, dev, pgpt->AlternateLBA,
					       &agpt, &aptes);
		if (!rc){
			printk(KERN_WARNING "Alternate GPT is invalid, using primary GPT.\n");
		}

		*gpt = pgpt;
		*ptes = pptes;
		if (agpt)  kfree(agpt);
		if (aptes) kfree(aptes);
		return 1;
	} /* if primary is valid */
	else {
		/* Primary GPT is bad, check the Alternate GPT */
		rc = IsGuidPartitionTableValid(hd, dev, lastlba,
					       &agpt, &aptes);
		if (rc) {
			/* Primary is bad, alternate is good.
			   Return values from the alternate and warn.
			*/
			printk(KERN_WARNING "Primary GPT is invalid, using alternate GPT.\n");
			*gpt = agpt;
			*ptes = aptes;
			return 1;
		}
		else {
			/* Primary is bad, alternate is bad, try "other"
			 * alternate.  This is necessary because if we
			 * have an odd-sized disk, user-space might
			 * have put the alternate in block lastlba-1.
			 */
			if (!(lastlba & 1)) {
				lastlba--;
				rc = IsGuidPartitionTableValid(hd, dev,
							       lastlba,
							       &agpt, &aptes);
				if (rc) {
					/* Primary is bad, alternate is good.
					 * Return values from the alternate
					 * and warn.
					 */
					printk("Primary GPT is invalid, using alternate GPT.\n");
					*gpt = agpt;
					*ptes = aptes;
					return 1;
				}
			}
		}
	}
	/* Both primary and alternate GPTs are bad.
	 * This isn't our disk, return 0.
	 */
	return 0;
}

#ifdef CONFIG_DEVFS_GUID
/* set_partition_guid */
/* Fill in the GUID field of the partition.
   It is set to NULL by register_disk before. */
static void set_partition_guid (struct gendisk *hd,
				const int minor,
				const efi_guid_t *guid)
{
	efi_guid_t *part_guid = hd->part[minor].guid;

	if (!guid || !hd) return;

	part_guid = kmalloc (sizeof (efi_guid_t), GFP_KERNEL);

        if (part_guid) {
		memcpy (part_guid, guid, sizeof (efi_guid_t));
	} else {
		printk (KERN_WARNING
                        "add_gpt_partitions: cannot allocate GUID memory!\n");
	};

	hd->part[minor].guid = part_guid;
}
#endif /* CONFIG_DEVFS_GUID */

/*
 * Create devices for each entry in the GUID Partition Table Entries.
 * The first block of each partition is a Legacy MBR.
 *
 * We do not create a Linux partition for GPT, but
 * only for the actual data partitions.
 * Returns:
 * -1 if unable to read the partition table
 *  0 if this isn't our partition table
 *  1 if successful
 *
 */

static int
add_gpt_partitions(struct gendisk *hd, kdev_t dev, int nextminor)
{
	GuidPartitionTableHeader_t *gpt = NULL;
	GuidPartitionEntry_t *ptes = NULL;
	u32 i, nummade=0;

	efi_guid_t unusedGuid = UNUSED_ENTRY_GUID;
#if CONFIG_BLK_DEV_MD
	efi_guid_t raidGuid = PARTITION_LINUX_RAID_GUID;
#endif

	if (!hd) return -1;

	if (!FindValidGPT(hd, dev, &gpt, &ptes) ||
	    !gpt || !ptes) {
		if (gpt) kfree(gpt);
		if (ptes) kfree(ptes);
		return 0;
	}

	debug_printk(efi_printk_level "GUID Partition Table is valid!  Yea!\n");

#ifdef CONFIG_DEVFS_GUID
	set_partition_guid (hd, nextminor - 1, &(gpt->DiskGUID));
#endif

	for (i = 0; i < gpt->NumberOfPartitionEntries &&
		     nummade < (hd->max_p - 1); i++) {
		if (!efi_guidcmp(unusedGuid, ptes[i].PartitionTypeGuid))
			continue;

		add_gd_partition(hd, nextminor, ptes[i].StartingLBA,
				 (ptes[i].EndingLBA-ptes[i].StartingLBA + 1));

#ifdef CONFIG_DEVFS_GUID
		set_partition_guid (hd, nextminor, &(ptes[i].UniquePartitionGuid));
#endif

		/* If there's this is a RAID volume, tell md */
#if CONFIG_BLK_DEV_MD
		if (!efi_guidcmp(raidGuid, ptes[i].PartitionTypeGuid)) {
			md_autodetect_dev(MKDEV(MAJOR(dev),nextminor));
		}
#endif
		nummade++;
		nextminor++;

	}
	kfree(ptes);
	kfree(gpt);
	printk("\n");
	return 1;

}


/*
 * efi_partition()
 *
 * If the first block on the disk is a legacy MBR,
 * it got handled already by msdos_partition().
 * If it's a Protective MBR, we'll handle it here.
 *
 * Returns:
 * -1 if unable to read the partition table
 *  0 if this isn't our partitoin table
 *  1 if successful
 *
 */

int
efi_partition(struct gendisk *hd, kdev_t dev,
	      unsigned long first_sector, int first_part_minor) {
	int hardblocksize = get_hardsect_size(dev);
	int orig_blksize_size = BLOCK_SIZE;
	int rc = 0;

	/* Need to change the block size that the block layer uses */
	if (blksize_size[MAJOR(dev)]){
		orig_blksize_size = blksize_size[MAJOR(dev)][MINOR(dev)];
	}

	if (orig_blksize_size != hardblocksize)
		set_blocksize(dev, hardblocksize);

	rc = add_gpt_partitions(hd, dev, first_part_minor);

	/* change back */
	if (orig_blksize_size != hardblocksize)
		set_blocksize(dev, orig_blksize_size);

	return rc;
}


/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */
