/************************************************************
 * EFI GUID Partition Table
 * Per Intel EFI Specification v0.99
 * http://developer.intel.com/technology/efi/efi.htm
 *
 * By Matt Domsch <Matt_Domsch@dell.com>  Fri Sep 22 22:15:56 CDT 2000  
 *   Copyright 2000 Dell Computer Corporation
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 ************************************************************/

#ifndef FS_PART_EFI_H_INCLUDED
#define FS_PART_EFI_H_INCLUDED

#include <linux/types.h>
#include <asm/efi.h>

#include <linux/config.h>
#include <linux/fs.h>
#include <linux/genhd.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/blk.h>

#define MSDOS_MBR_SIGNATURE 0xaa55
#define EFI_PMBR_OSTYPE_EFI 0xEF
#define EFI_PMBR_OSTYPE_EFI_GPT 0xEE

#define GUID_PT_BLOCK_SIZE 512

#define GUID_PT_HEADER_SIGNATURE 0x5452415020494645L
#define GUID_PT_HEADER_REVISION_V1 0x00010000
#define GUID_PT_HEADER_REVISION_V0_99 0x00000099
#define UNUSED_ENTRY_GUID    \
    ((efi_guid_t) { 0x00000000, 0x0000, 0x0000, { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }})
#define PARTITION_SYSTEM_GUID \
((efi_guid_t) { 0xC12A7328, 0xF81F, 0x11d2, { 0xBA, 0x4B, 0x00, 0xA0, 0xC9, 0x3E, 0xC9, 0x3B }})
#define LEGACY_MBR_PARTITION_GUID \
    ((efi_guid_t) { 0x024DEE41, 0x33E7, 0x11d3, { 0x9D, 0x69, 0x00, 0x08, 0xC7, 0x81, 0xF3, 0x9F }})
#define PARTITION_MSFT_RESERVED_GUID \
    ((efi_guid_t) { 0xE3C9E316, 0x0B5C, 0x4DB8, { 0x81, 0x7D, 0xF9, 0x2D, 0xF0, 0x02, 0x15, 0xAE }})
#define PARTITION_BASIC_DATA_GUID \
    ((efi_guid_t) { 0xEBD0A0A2, 0xB9E5, 0x4433, { 0x87, 0xC0, 0x68, 0xB6, 0xB7, 0x26, 0x99, 0xC7 }})
#define PARTITION_LINUX_RAID_GUID \
    ((efi_guid_t) { 0xa19d880f, 0x05fc, 0x4d3b, { 0xa0, 0x06, 0x74, 0x3f, 0x0f, 0x84, 0x91, 0x1e  }})
#define PARTITION_LINUX_SWAP_GUID \
    ((efi_guid_t) { 0x0657fd6d, 0xa4ab, 0x43c4, { 0x84, 0xe5, 0x09, 0x33, 0xc8, 0x4b, 0x4f, 0x4f  }})

typedef struct _GuidPartitionTableHeader_t {
  u64 Signature;
  u32 Revision;
  u32 HeaderSize;
  u32 HeaderCRC32;
  u32 Reserved1;
  u64 MyLBA;
  u64 AlternateLBA;
  u64 FirstUsableLBA;
  u64 LastUsableLBA;
  efi_guid_t DiskGUID;
  u64 PartitionEntryLBA;
  u32 NumberOfPartitionEntries;
  u32 SizeOfPartitionEntry;
  u32 PartitionEntryArrayCRC32;
  u8 Reserved2[GUID_PT_BLOCK_SIZE - 92];
} GuidPartitionTableHeader_t;

typedef struct _GuidPartitionEntryAttributes_t {
  __u64 RequiredToFunction:1;
  __u64 Reserved:63;
} GuidPartitionEntryAttributes_t;

typedef struct _GuidPartitionEntry_t {
  efi_guid_t PartitionTypeGuid;
  efi_guid_t UniquePartitionGuid;
  u64 StartingLBA;
  u64 EndingLBA;
  GuidPartitionEntryAttributes_t Attributes;
  efi_char16_t PartitionName[72/sizeof(efi_char16_t)];
} GuidPartitionEntry_t;



typedef struct _PartitionRecord_t {
  u8 BootIndicator;  /* Not used by EFI firmware. Set to 0x80 to indicate that this
                        is the bootable legacy partition. */
  u8 StartHead;      /* Start of partition in CHS address, not used by EFI firmware. */
  u8 StartSector;    /* Start of partition in CHS address, not used by EFI firmware. */
  u8 StartTrack;     /* Start of partition in CHS address, not used by EFI firmware. */
  u8 OSType;         /* OS type. A value of 0xEF defines an EFI system partition.
                        Other values are reserved for legacy operating systems, and
                        allocated independently of the EFI specification. */
  u8 EndHead;        /* End of partition in CHS address, not used by EFI firmware. */
  u8 EndSector;      /* End of partition in CHS address, not used by EFI firmware. */
  u8 EndTrack;       /* End of partition in CHS address, not used by EFI firmware. */
  u32 StartingLBA;   /* Starting LBA address of the partition on the disk. Used by
                        EFI firmware to define the start of the partition. */
  u32 SizeInLBA;     /* Size of partition in LBA. Used by EFI firmware to determine
                        the size of the partition. */
} PartitionRecord_t;

typedef struct _LegacyMBR_t {
  u8 BootCode[440];
  u32 UniqueMBRSignature;
  u16 Unknown;
  PartitionRecord_t PartitionRecord[4];
  u16 Signature;
} __attribute__ ((packed)) LegacyMBR_t;



#define EFI_GPT_PRIMARY_PARTITION_TABLE_LBA 1

/* Functions */
extern int
efi_partition(struct gendisk *hd, kdev_t dev,
              unsigned long first_sector, int first_part_minor);




#endif

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * --------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4 
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */
